"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.BlockEventHandler = void 0;
const lisk_sdk_1 = require("lisk-sdk");
const constants_1 = require("./constants");
const ccu_handler_1 = require("./ccu_handler");
class BlockEventHandler {
    constructor(config) {
        this._isReceivingChainRegistered = false;
        this._lastSentCCUTxID = '';
        this._ownChainID = config.ownChainID;
        this._ccuSaveLimit = config.ccuSaveLimit;
        this._receivingChainFinalizedHeight = 0;
        this._receivingChainID = config.receivingChainID;
        this._isReceivingChainMainchain = !(0, lisk_sdk_1.getMainchainID)(this._ownChainID).equals(this._ownChainID);
        this._ccuHandler = new ccu_handler_1.CCUHandler({
            maxCCUSize: config.maxCCUSize,
            ownChainID: config.ownChainID,
            receivingChainID: config.receivingChainID,
            registrationHeight: config.registrationHeight,
            ccuFee: config.ccuFee,
            isSaveCCU: config.isSaveCCU,
        });
    }
    async load(args) {
        this._logger = args.logger;
        this._db = args.db;
        this._sendingChainAPIClient = args.sendingChainAPIClient;
        this._receivingChainAPIClient = args.receivingChainAPIClient;
        this._heightToDeleteIndex = new Map();
        this._interoperabilityMetadata = await this._sendingChainAPIClient.getMetadataByModuleName(lisk_sdk_1.MODULE_NAME_INTEROPERABILITY);
        this._ccuHandler.load({
            db: args.db,
            lastCertificate: this._lastCertificate,
            logger: args.logger,
            receivingChainAPIClient: args.receivingChainAPIClient,
            sendingChainAPIClient: args.sendingChainAPIClient,
            interoperabilityMetadata: this._interoperabilityMetadata,
        });
        await this._receivingChainAPIClient.connect();
        this._lastIncludedCCMOnReceivingChain = await this._db.getLastSentCCM();
        this._sendingChainAPIClient.subscribe('chain_newBlock', async (data) => {
            try {
                await this._handleNewBlock(data);
            }
            catch (error) {
                this._logger.error({ err: error }, 'Failed to handle new block.');
            }
        });
        this._sendingChainAPIClient.subscribe('chain_deleteBlock', async (data) => {
            try {
                await this._deleteBlockHandler(data);
            }
            catch (error) {
                this._logger.error({ err: error }, 'Failed to handle delete block.');
            }
        });
        this._initializeReceivingChainClient().catch(this._logger.error);
    }
    async _handleNewBlock(data) {
        const { blockHeader: receivedBlock } = data;
        const newBlockHeader = lisk_sdk_1.chain.BlockHeader.fromJSON(receivedBlock).toObject();
        await this._saveOnNewBlock(newBlockHeader);
        const nodeInfo = await this._sendingChainAPIClient.getNodeInfo();
        if (nodeInfo.syncing) {
            this._logger.debug('No CCU generation is possible as the node is syncing.');
            return;
        }
        let chainAccount;
        try {
            chainAccount = await this._receivingChainAPIClient.getChainAccount(this._ownChainID);
        }
        catch (error) {
            await this._initializeReceivingChainClient();
            this._logger.error({ err: error }, 'Error occurred while accessing receivingChainAPIClient but all data is saved on new block.');
            return;
        }
        if (!chainAccount || (chainAccount && !chainAccount.lastCertificate)) {
            this._logger.info('Sending chain is not registered to the receiving chain yet and has no chain data.');
            return;
        }
        this._lastCertificate = chainAccount.lastCertificate;
        this._logger.info(`Last certificate value has been set with height ${this._lastCertificate.height}`);
        const numOfBlocksSinceLastCertificate = newBlockHeader.height - this._lastCertificate.height;
        if (this._ccuFrequency > numOfBlocksSinceLastCertificate) {
            this._logger.info({
                ccuFrequency: this._ccuFrequency,
                nextPossibleCCUHeight: this._ccuFrequency - numOfBlocksSinceLastCertificate,
            }, 'No attempt to create CCU either due to provided ccuFrequency');
            return;
        }
        if (!this._isReceivingChainRegistered) {
            const receivingChainAccount = await this._sendingChainAPIClient.getChainAccount(this._receivingChainID);
            if (!receivingChainAccount) {
                this._logger.info('Receiving chain is not registered on the sending chain yet and has no chain data.');
                return;
            }
            this._isReceivingChainRegistered = true;
        }
        let computedCCUParams;
        if (this._lastSentCCUTxID === '') {
            computedCCUParams = await this._ccuHandler.computeCCU(this._lastCertificate, this._lastIncludedCCMOnReceivingChain);
        }
        else {
            this._logger.info(`Still pending CCU on the receiving CCU with tx ID ${this._lastSentCCUTxID}`);
            return;
        }
        if (!computedCCUParams) {
            this._logger.info(`No CCU params were generated for block height ${newBlockHeader.height}`);
            return;
        }
        const ccuSubmitResult = await this._ccuHandler.submitCCU(computedCCUParams.ccuParams, this._lastSentCCUTxID);
        if (ccuSubmitResult) {
            this._lastSentCCUTxID = ccuSubmitResult;
            this._sentCCUTxTimeout = setTimeout(() => {
                this._lastSentCCUTxID = '';
                clearTimeout(this._sentCCUTxTimeout);
            }, constants_1.DEFAULT_SENT_CCU_TIMEOUT);
            if (computedCCUParams.lastCCMToBeSent) {
                this._lastSentCCM = computedCCUParams.lastCCMToBeSent;
                this._logger.info(`Last sent CCM with nonce ${this._lastSentCCM.nonce.toString()} is saved.`);
            }
            this._logger.info(`CCU transaction was successfully sent with Tx ID ${ccuSubmitResult}.`);
            return;
        }
        this._logger.info(`Last sent CCU tx with ID ${this._lastSentCCUTxID} was not yet included in the receiving chain.`);
    }
    async _saveOnNewBlock(newBlockHeader) {
        var _a, _b;
        await this._db.saveToDBOnNewBlock(newBlockHeader);
        const events = await this._sendingChainAPIClient.getEvents(newBlockHeader.height);
        const ccmsFromEvents = [];
        ccmInEventsCheck: if (events && events.length > 0) {
            const ccmSendSuccessEvents = events.filter(eventAttr => eventAttr.name === constants_1.CCM_SEND_SUCCESS && eventAttr.module === lisk_sdk_1.MODULE_NAME_INTEROPERABILITY);
            const ccmProcessedEvents = events.filter(eventAttr => eventAttr.name === constants_1.CCM_PROCESSED && eventAttr.module === lisk_sdk_1.MODULE_NAME_INTEROPERABILITY);
            if (ccmSendSuccessEvents.length === 0 && ccmProcessedEvents.length === 0) {
                break ccmInEventsCheck;
            }
            if (ccmSendSuccessEvents.length > 0) {
                const ccmSendSuccessEventInfo = this._interoperabilityMetadata.events.filter(e => e.name === constants_1.CCM_SEND_SUCCESS);
                if (!((_a = ccmSendSuccessEventInfo === null || ccmSendSuccessEventInfo === void 0 ? void 0 : ccmSendSuccessEventInfo[0]) === null || _a === void 0 ? void 0 : _a.data)) {
                    throw new Error('No schema found for "ccmSendSuccess" event data.');
                }
                for (const e of ccmSendSuccessEvents) {
                    const eventData = lisk_sdk_1.codec.decode(ccmSendSuccessEventInfo[0].data, Buffer.from(e.data, 'hex'));
                    ccmsFromEvents.push(eventData.ccm);
                }
            }
            if (ccmProcessedEvents.length > 0) {
                const ccmProcessedEventInfo = this._interoperabilityMetadata.events.filter(e => e.name === constants_1.CCM_PROCESSED);
                if (!((_b = ccmProcessedEventInfo === null || ccmProcessedEventInfo === void 0 ? void 0 : ccmProcessedEventInfo[0]) === null || _b === void 0 ? void 0 : _b.data)) {
                    throw new Error('No schema found for "ccmProcessed" event data.');
                }
                for (const e of ccmProcessedEvents) {
                    const eventData = lisk_sdk_1.codec.decode(ccmProcessedEventInfo[0].data, Buffer.from(e.data, 'hex'));
                    if (eventData.result === 1) {
                        ccmsFromEvents.push(eventData.ccm);
                    }
                }
            }
        }
        await this._db.setCCMsByHeight(ccmsFromEvents
            .map(ccm => ({ ...ccm, height: newBlockHeader.height }))
            .filter(ccm => this._isReceivingChainMainchain || ccm.receivingChainID.equals(this._receivingChainID)), newBlockHeader.height);
        const validatorsData = await this._sendingChainAPIClient.getBFTParametersAtHeight(newBlockHeader.height);
        await this._db.setValidatorsDataByHash(validatorsData.validatorsHash, { ...validatorsData, height: newBlockHeader.height }, newBlockHeader.height);
    }
    async _initializeReceivingChainClient() {
        try {
            await this._receivingChainAPIClient.connect();
            this._receivingChainAPIClient.subscribe('chain_newBlock', async (data) => this._newBlockReceivingChainHandler(data));
        }
        catch (error) {
            this._logger.error(error, 'Not able to connect to receivingChainAPIClient. Trying again on next new block.');
        }
    }
    async _newBlockReceivingChainHandler(_) {
        var _a;
        try {
            const { finalizedHeight, syncing } = await this._receivingChainAPIClient.getNodeInfo();
            if (syncing) {
                this._logger.debug('Receiving chain is syncing.');
                return;
            }
            this._receivingChainFinalizedHeight = finalizedHeight;
            const channelDataOnReceivingChain = await this._receivingChainAPIClient.getChannelAccount(this._ownChainID);
            if (!channelDataOnReceivingChain) {
                throw new Error('No channel data available on receiving chain.');
            }
            const chainAccount = await this._receivingChainAPIClient.getChainAccount(this._ownChainID);
            if (!chainAccount) {
                throw new Error('No chain data available on receiving chain.');
            }
            this._heightToDeleteIndex.set(finalizedHeight, {
                inboxSize: channelDataOnReceivingChain.inbox.size,
                lastCertificateHeight: (_a = chainAccount.lastCertificate) === null || _a === void 0 ? void 0 : _a.height,
            });
            if (this._lastSentCCUTxID !== '') {
                try {
                    await this._receivingChainAPIClient.getTransactionByID(this._lastSentCCUTxID);
                    this._logger.info(`CCU transaction with ${this._lastSentCCUTxID} was included on the receiving chain`);
                    this._lastSentCCUTxID = '';
                    clearTimeout(this._sentCCUTxTimeout);
                    if (this._lastSentCCM) {
                        this._lastIncludedCCMOnReceivingChain = this._lastSentCCM;
                        await this._db.setLastSentCCM(this._lastIncludedCCMOnReceivingChain);
                    }
                }
                catch (error) {
                    throw new Error(`Failed to get transaction with ID ${this._lastSentCCUTxID}`);
                }
            }
            await this._cleanup();
        }
        catch (error) {
            this._logger.debug(error, 'Error occured while receiving block from receiving chain.');
        }
    }
    async _cleanup() {
        var _a;
        if (this._ccuSaveLimit !== -1) {
            const { list: listOfCCUs, total } = await this._db.getListOfCCUs();
            if (total > this._ccuSaveLimit) {
                for (let i = total - 1; i >= this._ccuSaveLimit; i -= 1) {
                    await this._db.deleteCCUTransaction(Buffer.from((_a = listOfCCUs[i]) === null || _a === void 0 ? void 0 : _a.id, 'hex'));
                }
            }
        }
        let finalizedInfoAtHeight = this._heightToDeleteIndex.get(this._receivingChainFinalizedHeight);
        if (!finalizedInfoAtHeight) {
            for (let i = 1; i < this._heightToDeleteIndex.size; i += 1) {
                if (this._heightToDeleteIndex.get(this._receivingChainFinalizedHeight - i)) {
                    finalizedInfoAtHeight = this._heightToDeleteIndex.get(this._receivingChainFinalizedHeight - i);
                    break;
                }
            }
        }
        const endDeletionHeightByLastCertificate = finalizedInfoAtHeight
            ? finalizedInfoAtHeight.lastCertificateHeight
            : 0;
        if (this._lastCertificate.height > 0) {
            await this._db.deleteCCMsBetweenHeight(this._lastDeletionHeight, endDeletionHeightByLastCertificate - 1);
            await this._db.deleteBlockHeadersBetweenHeight(this._lastDeletionHeight, endDeletionHeightByLastCertificate - 1);
            await this._db.deleteAggregateCommitsBetweenHeight(this._lastDeletionHeight, endDeletionHeightByLastCertificate - 1);
            await this._db.deleteValidatorsHashBetweenHeights(this._lastDeletionHeight, endDeletionHeightByLastCertificate - 1);
            this._lastDeletionHeight = endDeletionHeightByLastCertificate;
            this._logger.debug(`Deleted data on cleanup between heights 1 and ${endDeletionHeightByLastCertificate}`);
        }
        this._heightToDeleteIndex.forEach((_, key) => {
            if (key < this._receivingChainFinalizedHeight) {
                this._heightToDeleteIndex.delete(key);
            }
        });
    }
    async _deleteBlockHandler(data) {
        const { blockHeader: receivedBlock } = data;
        const deletedBlockHeader = lisk_sdk_1.chain.BlockHeader.fromJSON(receivedBlock).toObject();
        await this._db.deleteCCMsByHeight(deletedBlockHeader.height);
        await this._db.deleteBlockHeaderByHeight(deletedBlockHeader.height);
        await this._db.deleteAggregateCommitByHeight(deletedBlockHeader.height);
        await this._db.deleteValidatorsHashByHeight(deletedBlockHeader.height);
    }
}
exports.BlockEventHandler = BlockEventHandler;
//# sourceMappingURL=block_event_handler.js.map