"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.CCUHandler = void 0;
const lisk_sdk_1 = require("lisk-sdk");
const certificate_generation_1 = require("./certificate_generation");
const constants_1 = require("./constants");
const inbox_update_1 = require("./inbox_update");
const active_validators_update_1 = require("./active_validators_update");
class CCUHandler {
    constructor(config) {
        this._registrationHeight = config.registrationHeight;
        this._ownChainID = config.ownChainID;
        this._receivingChainID = config.receivingChainID;
        this._maxCCUSize = config.maxCCUSize;
        this._ccuFee = config.ccuFee;
        this._isSaveCCU = config.isSaveCCU;
        this._isReceivingChainMainchain = !(0, lisk_sdk_1.getMainchainID)(this._ownChainID).equals(this._ownChainID);
    }
    load(args) {
        this._logger = args.logger;
        this._db = args.db;
        this._sendingChainAPIClient = args.sendingChainAPIClient;
        this._receivingChainAPIClient = args.receivingChainAPIClient;
        this._lastCertificate = args.lastCertificate;
        this._interoperabilityMetadata = args.interoperabilityMetadata;
        const store = this._interoperabilityMetadata.stores.find(s => s.data.$id === '/modules/interoperability/outbox');
        this._outboxKeyForInclusionProof = Buffer.concat([
            Buffer.from(store === null || store === void 0 ? void 0 : store.key, 'hex'),
            lisk_sdk_1.cryptography.utils.hash(this._receivingChainID),
        ]);
    }
    async computeCCU(lastCertificate, lastIncludedCCM) {
        this._lastCertificate = lastCertificate;
        const newCertificate = await this._findCertificate();
        if (!newCertificate && this._lastCertificate.height === 0) {
            return undefined;
        }
        const lastSentCCM = lastIncludedCCM !== null && lastIncludedCCM !== void 0 ? lastIncludedCCM : {
            nonce: constants_1.DEFAULT_LAST_CCM_SENT_NONCE,
            height: this._lastCertificate.height,
        };
        const ccmsRange = await this._db.getCCMsBetweenHeights(lastSentCCM.height, newCertificate ? newCertificate.height : this._lastCertificate.height);
        const channelDataOnReceivingChain = await this._receivingChainAPIClient.getChannelAccount(this._ownChainID);
        if (!channelDataOnReceivingChain) {
            return undefined;
        }
        const channelDataOnSendingChain = await this._sendingChainAPIClient.getChannelAccount(this._receivingChainID);
        if (!channelDataOnSendingChain) {
            return undefined;
        }
        const { crossChainMessages, lastCCMToBeSent, messageWitnessHashes } = (0, inbox_update_1.calculateMessageWitnesses)(channelDataOnReceivingChain.inbox.size, channelDataOnSendingChain === null || channelDataOnSendingChain === void 0 ? void 0 : channelDataOnSendingChain.outbox.size, lastSentCCM, ccmsRange, this._maxCCUSize);
        let activeValidatorsUpdate = {
            blsKeysUpdate: [],
            bftWeightsUpdate: [],
            bftWeightsUpdateBitmap: lisk_sdk_1.EMPTY_BYTES,
        };
        let certificate = lisk_sdk_1.EMPTY_BYTES;
        let certificateThreshold;
        let outboxRootWitness;
        if (!newCertificate) {
            if (crossChainMessages.length === 0) {
                this._logger.info('CCU cant be created as there are no pending CCMs for the last certificate.');
                return undefined;
            }
            outboxRootWitness = {
                bitmap: lisk_sdk_1.EMPTY_BYTES,
                siblingHashes: [],
            };
            const validatorsDataAtLastCertificate = await this._db.getValidatorsDataByHash(this._lastCertificate.validatorsHash);
            certificateThreshold = validatorsDataAtLastCertificate === null || validatorsDataAtLastCertificate === void 0 ? void 0 : validatorsDataAtLastCertificate.certificateThreshold;
            return {
                ccuParams: {
                    sendingChainID: this._ownChainID,
                    activeValidatorsUpdate,
                    certificate,
                    certificateThreshold,
                    inboxUpdate: {
                        crossChainMessages,
                        messageWitnessHashes,
                        outboxRootWitness,
                    },
                },
                lastCCMToBeSent,
            };
        }
        const validatorsDataAtLastCertificate = await this._db.getValidatorsDataByHash(this._lastCertificate.validatorsHash);
        if (!validatorsDataAtLastCertificate) {
            throw new Error(`No validators data at last certificate with hash at ${this._lastCertificate.validatorsHash.toString('hex')}`);
        }
        if (!this._lastCertificate.validatorsHash.equals(newCertificate.validatorsHash)) {
            const validatorsDataAtNewCertificate = await this._db.getValidatorsDataByHash(newCertificate.validatorsHash);
            if (!validatorsDataAtNewCertificate) {
                throw new Error(`No validators data at new certificate with hash at ${newCertificate.validatorsHash.toString('hex')}`);
            }
            const validatorsUpdateResult = (0, active_validators_update_1.calculateActiveValidatorsUpdate)(validatorsDataAtLastCertificate, validatorsDataAtNewCertificate);
            activeValidatorsUpdate = validatorsUpdateResult.activeValidatorsUpdate;
            certificateThreshold = validatorsUpdateResult.certificateThreshold;
        }
        else {
            certificateThreshold = validatorsDataAtLastCertificate === null || validatorsDataAtLastCertificate === void 0 ? void 0 : validatorsDataAtLastCertificate.certificateThreshold;
        }
        if (crossChainMessages.length === 0) {
            outboxRootWitness = {
                bitmap: lisk_sdk_1.EMPTY_BYTES,
                siblingHashes: [],
            };
        }
        else {
            const inclusionProofs = await this._sendingChainAPIClient.getSavedInclusionProofAtHeight(newCertificate.height);
            const foundInclusionProof = inclusionProofs.proof.queries.find(q => q.key.equals(this._outboxKeyForInclusionProof));
            if (!foundInclusionProof) {
                throw new Error(`No inclusion proof was found for key ${this._outboxKeyForInclusionProof.toString('hex')}`);
            }
            outboxRootWitness = {
                bitmap: foundInclusionProof.bitmap,
                siblingHashes: inclusionProofs.proof.siblingHashes,
            };
        }
        certificate = lisk_sdk_1.codec.encode(lisk_sdk_1.certificateSchema, newCertificate);
        return {
            ccuParams: {
                sendingChainID: this._ownChainID,
                activeValidatorsUpdate,
                certificate,
                certificateThreshold,
                inboxUpdate: {
                    crossChainMessages,
                    messageWitnessHashes,
                    outboxRootWitness,
                },
            },
            lastCCMToBeSent,
        };
    }
    async submitCCU(ccuParams, lastSentCCUTxID) {
        if (!this._db.privateKey) {
            throw new Error('There is no key enabled to submit CCU.');
        }
        const { syncing } = await this._receivingChainAPIClient.getNodeInfo();
        if (syncing) {
            throw new Error('Receiving node is syncing.');
        }
        const relayerPublicKey = lisk_sdk_1.cryptography.ed.getPublicKeyFromPrivateKey(this._db.privateKey);
        const targetCommand = this._isReceivingChainMainchain
            ? constants_1.COMMAND_NAME_SUBMIT_MAINCHAIN_CCU
            : constants_1.COMMAND_NAME_SUBMIT_SIDECHAIN_CCU;
        const nonce = await this._receivingChainAPIClient.getAuthAccountNonceFromPublicKey(relayerPublicKey);
        const txWithoutFee = {
            module: lisk_sdk_1.MODULE_NAME_INTEROPERABILITY,
            command: targetCommand,
            nonce: BigInt(nonce),
            senderPublicKey: relayerPublicKey,
            params: lisk_sdk_1.codec.encode(lisk_sdk_1.ccuParamsSchema, ccuParams),
            signatures: [],
        };
        const tx = new lisk_sdk_1.Transaction({
            ...txWithoutFee,
            fee: await this._getCcuFee({
                ...txWithoutFee,
                params: ccuParams,
            }),
        });
        tx.sign(this._receivingChainID, this._db.privateKey);
        if (tx.id.equals(Buffer.from(lastSentCCUTxID, 'hex'))) {
            return undefined;
        }
        let result;
        if (this._isSaveCCU) {
            result = { transactionId: tx.id.toString('hex') };
        }
        else {
            result = await this._receivingChainAPIClient.postTransaction(tx.getBytes());
        }
        await this._db.setCCUTransaction(tx.toObject());
        return result.transactionId;
    }
    async _findCertificate() {
        if (this._lastCertificate.height === 0) {
            const aggreggateCommits = await this._db.getAggregateCommitBetweenHeights(this._registrationHeight, 1000);
            for (const aggregateCommit of aggreggateCommits) {
                const blockHeader = await this._db.getBlockHeaderByHeight(aggregateCommit.height);
                if (!blockHeader) {
                    continue;
                }
                const firstCertificate = (0, certificate_generation_1.getCertificateFromAggregateCommitByBlockHeader)(aggregateCommit, blockHeader);
                return firstCertificate;
            }
            return undefined;
        }
        const bftHeights = await this._sendingChainAPIClient.getBFTHeights();
        return (0, certificate_generation_1.getNextCertificateFromAggregateCommits)(this._db, bftHeights, this._lastCertificate);
    }
    async _getCcuFee(tx) {
        var _a;
        let additionalFee = BigInt(0);
        const userBalance = await this._receivingChainAPIClient.hasUserTokenAccount(lisk_sdk_1.cryptography.address.getLisk32AddressFromAddress(lisk_sdk_1.cryptography.address.getAddressFromPublicKey(tx.senderPublicKey)));
        if (!userBalance.exists) {
            const fee = await this._receivingChainAPIClient.getTokenInitializationFee();
            additionalFee += BigInt(fee.userAccount);
        }
        const ccuFee = BigInt((_a = this._ccuFee) !== null && _a !== void 0 ? _a : '0') + additionalFee;
        const computedMinFee = lisk_sdk_1.transactions.computeMinFee(tx, lisk_sdk_1.ccuParamsSchema, {
            additionalFee,
        });
        if (ccuFee > computedMinFee) {
            return ccuFee;
        }
        return computedMinFee;
    }
}
exports.CCUHandler = CCUHandler;
//# sourceMappingURL=ccu_handler.js.map