"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.getNextCertificateFromAggregateCommits = exports.checkChainOfTrust = exports.getCertificateFromAggregateCommitByBlockHeader = void 0;
const lisk_sdk_1 = require("lisk-sdk");
const getCertificateFromAggregateCommitByBlockHeader = (aggregateCommit, blockHeader) => ({
    ...(0, lisk_sdk_1.computeUnsignedCertificateFromBlockHeader)(new lisk_sdk_1.chain.BlockHeader(blockHeader)),
    aggregationBits: aggregateCommit.aggregationBits,
    signature: aggregateCommit.certificateSignature,
});
exports.getCertificateFromAggregateCommitByBlockHeader = getCertificateFromAggregateCommitByBlockHeader;
const checkChainOfTrust = async (lastValidatorsHash, blsKeyToBFTWeight, lastCertificateThreshold, aggregateCommit, db) => {
    const blockHeader = await db.getBlockHeaderByHeight(aggregateCommit.height - 1);
    if (!blockHeader) {
        throw new Error(`No block header found for the given the previous height ${aggregateCommit.height - 1} of aggregate commit at height ${aggregateCommit.height} when calling checkChainOfTrust.`);
    }
    if (lastValidatorsHash.equals(blockHeader.validatorsHash)) {
        return true;
    }
    let aggregateBFTWeight = BigInt(0);
    const validatorData = await db.getValidatorsDataByHash(blockHeader.validatorsHash);
    if (!validatorData) {
        throw new Error(`No validators data found for the given validatorsHash ${blockHeader.validatorsHash.toString('hex')}.`);
    }
    for (let i = 0; i < validatorData.validators.length; i += 1) {
        if (aggregateCommit.aggregationBits.toString('hex')[i] === '1') {
            const blsKey = validatorData.validators[i].blsKey.toString('hex');
            if (!blsKeyToBFTWeight[blsKey]) {
                return false;
            }
            aggregateBFTWeight += blsKeyToBFTWeight[blsKey];
        }
    }
    return aggregateBFTWeight >= lastCertificateThreshold;
};
exports.checkChainOfTrust = checkChainOfTrust;
const getNextCertificateFromAggregateCommits = async (db, bftHeights, lastCertificate) => {
    const blockHeaderAtLastCertifiedHeight = await db.getBlockHeaderByHeight(lastCertificate.height);
    if (!blockHeaderAtLastCertifiedHeight) {
        throw new Error(`No block header found for the last certified height ${lastCertificate.height}.`);
    }
    const validatorDataAtLastCertifiedHeight = await db.getValidatorsDataByHash(blockHeaderAtLastCertifiedHeight === null || blockHeaderAtLastCertifiedHeight === void 0 ? void 0 : blockHeaderAtLastCertifiedHeight.validatorsHash);
    if (!validatorDataAtLastCertifiedHeight) {
        throw new Error(`No validatorsHash preimage data present for the given validatorsHash ${blockHeaderAtLastCertifiedHeight === null || blockHeaderAtLastCertifiedHeight === void 0 ? void 0 : blockHeaderAtLastCertifiedHeight.validatorsHash.toString('hex')}.`);
    }
    const blsKeyToBFTWeight = {};
    for (const validator of validatorDataAtLastCertifiedHeight.validators) {
        blsKeyToBFTWeight[validator.blsKey.toString('hex')] = validator.bftWeight;
    }
    let height = bftHeights.maxHeightCertified;
    while (height > lastCertificate.height) {
        const aggregateCommitAtHeight = await db.getAggregateCommitByHeight(height);
        if (aggregateCommitAtHeight !== undefined) {
            const valid = await (0, exports.checkChainOfTrust)(blockHeaderAtLastCertifiedHeight.validatorsHash, blsKeyToBFTWeight, validatorDataAtLastCertifiedHeight.certificateThreshold, aggregateCommitAtHeight, db);
            if (valid) {
                const blockHeaderAtAggregateCommitHeight = await db.getBlockHeaderByHeight(aggregateCommitAtHeight.height);
                if (!blockHeaderAtAggregateCommitHeight) {
                    throw new Error(`Block header not found for the given aggregate commit for height: ${aggregateCommitAtHeight.height}`);
                }
                return (0, exports.getCertificateFromAggregateCommitByBlockHeader)(aggregateCommitAtHeight, blockHeaderAtAggregateCommitHeight);
            }
        }
        height -= 1;
    }
    return undefined;
};
exports.getNextCertificateFromAggregateCommits = getNextCertificateFromAggregateCommits;
//# sourceMappingURL=certificate_generation.js.map