"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.ChainAPIClient = void 0;
const lisk_sdk_1 = require("lisk-sdk");
const utils_1 = require("./utils");
const { address } = lisk_sdk_1.cryptography;
class ChainAPIClient {
    constructor(config) {
        this._config = config;
    }
    async connect(client) {
        if (client) {
            this._client = client;
            return;
        }
        if (!this._config.ipcPath && !this._config.wsConnectionString) {
            throw new Error('IPC path and WS url are undefined in the configuration.');
        }
        if (this._config.ipcPath) {
            this._client = await lisk_sdk_1.apiClient.createIPCClient(this._config.ipcPath);
        }
        else if (this._config.wsConnectionString) {
            this._client = await lisk_sdk_1.apiClient.createWSClient(this._config.wsConnectionString);
        }
        this.chainID = Buffer.from((await this.getNodeInfo()).chainID, 'hex');
    }
    async disconnect() {
        await this._client.disconnect();
    }
    subscribe(eventName, cb) {
        var _a;
        (_a = this._client) === null || _a === void 0 ? void 0 : _a.subscribe(eventName, cb);
    }
    async postTransaction(txBytes) {
        var _a;
        const result = await ((_a = this._client) === null || _a === void 0 ? void 0 : _a.invoke('txpool_postTransaction', {
            transaction: txBytes.toString('hex'),
        }));
        return result;
    }
    async getTransactionByID(id) {
        var _a;
        const result = await ((_a = this._client) === null || _a === void 0 ? void 0 : _a.invoke('chain_getTransactionByID', {
            id,
        }));
        return result;
    }
    async getAuthAccountNonceFromPublicKey(publicKey) {
        return (await this._client.invoke('auth_getAuthAccount', {
            address: address.getLisk32AddressFromPublicKey(publicKey),
        })).nonce;
    }
    async getNodeInfo() {
        return this._client.node.getNodeInfo();
    }
    async getChannelAccount(chainID) {
        const channelAccount = await this._client.invoke('interoperability_getChannel', {
            chainID: chainID.toString('hex'),
        });
        if (!channelAccount || (channelAccount === null || channelAccount === void 0 ? void 0 : channelAccount.inbox) === undefined) {
            return undefined;
        }
        return (0, utils_1.channelDataJSONToObj)(channelAccount);
    }
    async getChainAccount(chainID) {
        const chainAccount = await this._client.invoke('interoperability_getChainAccount', {
            chainID: chainID.toString('hex'),
        });
        if (!chainAccount || (chainAccount === null || chainAccount === void 0 ? void 0 : chainAccount.lastCertificate) === undefined) {
            return undefined;
        }
        return (0, utils_1.chainAccountDataJSONToObj)(chainAccount);
    }
    async hasUserTokenAccount(userAddress) {
        return this._client.invoke('token_hasUserAccount', {
            address: userAddress,
            tokenID: `${(0, utils_1.getTokenIDLSK)(this.chainID).toString('hex')}`,
        });
    }
    async getTokenInitializationFee() {
        return this._client.invoke('token_getInitializationFees');
    }
    async getBFTHeights() {
        return this._client.invoke('consensus_getBFTHeights');
    }
    async getEvents(height) {
        return this._client.invoke('chain_getEvents', { height });
    }
    async getMetadataByModuleName(moduleName) {
        const { modules: modulesMetadata } = await this._client.invoke('system_getMetadata');
        const moduleMetadata = modulesMetadata.find(m => m.name === moduleName);
        if (!moduleMetadata) {
            throw new Error(`No metadata found for ${moduleName} module.`);
        }
        return moduleMetadata;
    }
    async getInclusionProof(queryKeys) {
        return (0, utils_1.proveResponseJSONToObj)(await this._client.invoke('state_prove', {
            queryKeys: [...queryKeys].map(k => k.toString('hex')),
        }));
    }
    async getSavedInclusionProofAtHeight(height) {
        return (0, utils_1.proveResponseJSONToObj)(await this._client.invoke('chain_getInclusionProofsAtHeight', {
            height,
        }));
    }
    async getBFTParametersAtHeight(height) {
        return (0, utils_1.bftParametersJSONToObj)(await this._client.invoke('consensus_getBFTParametersActiveValidators', {
            height,
        }));
    }
}
exports.ChainAPIClient = ChainAPIClient;
//# sourceMappingURL=chain_api_client.js.map