"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.ChainConnectorPlugin = void 0;
const lisk_sdk_1 = require("lisk-sdk");
const constants_1 = require("./constants");
const schemas_1 = require("./schemas");
const chain_api_client_1 = require("./chain_api_client");
const block_event_handler_1 = require("./block_event_handler");
const db_1 = require("./db");
const endpoint_1 = require("./endpoint");
class ChainConnectorPlugin extends lisk_sdk_1.BasePlugin {
    constructor() {
        super(...arguments);
        this.endpoint = new endpoint_1.ChainConnectorEndpoint();
        this.configSchema = schemas_1.configSchema;
        this._chainConnectorDB = new db_1.ChainConnectorDB();
    }
    get nodeModulePath() {
        return __filename;
    }
    async init(context) {
        await super.init(context);
        if (this.config.maxCCUSize > constants_1.CCU_TOTAL_CCM_SIZE) {
            throw new Error(`maxCCUSize cannot be greater than ${constants_1.CCU_TOTAL_CCM_SIZE} bytes.`);
        }
        this._receivingChainID = Buffer.from(this.config.receivingChainID, 'hex');
        this._blockEventHandler = new block_event_handler_1.BlockEventHandler({
            maxCCUSize: this.config.maxCCUSize,
            ownChainID: Buffer.from(this.appConfig.genesis.chainID, 'hex'),
            receivingChainID: Buffer.from(this.config.receivingChainID, 'hex'),
            registrationHeight: this.config.registrationHeight,
            ccuFee: this.config.ccuFee,
            isSaveCCU: this.config.isSaveCCU,
            ccuSaveLimit: this.config.ccuSaveLimit,
        });
        this._sendingChainClient = new chain_api_client_1.ChainAPIClient({
            ipcPath: this.appConfig.system.dataPath,
            logger: this.logger,
        });
    }
    async load() {
        await this._chainConnectorDB.load(this.dataPath);
        this.endpoint.load(this.config.encryptedPrivateKey, this._chainConnectorDB);
        await this._sendingChainClient.connect(this.apiClient);
        this._ownChainID = Buffer.from(this.appConfig.genesis.chainID, 'hex');
        if (this._receivingChainID[0] !== this._ownChainID[0]) {
            throw new Error('Receiving Chain ID network does not match the sending chain network.');
        }
        this._receivingChainClient = new chain_api_client_1.ChainAPIClient({
            logger: this.logger,
            ipcPath: this.config.receivingChainIPCPath,
            wsConnectionString: this.config.receivingChainWsURL,
        });
        await this._blockEventHandler.load({
            db: this._chainConnectorDB,
            logger: this.logger,
            receivingChainAPIClient: this._receivingChainClient,
            sendingChainAPIClient: this._sendingChainClient,
        });
    }
    async unload() {
        if (this._receivingChainClient) {
            await this._receivingChainClient.disconnect();
        }
        if (this._sendingChainClient) {
            await this._sendingChainClient.disconnect();
        }
        this._chainConnectorDB.close();
    }
}
exports.ChainConnectorPlugin = ChainConnectorPlugin;
//# sourceMappingURL=chain_connector_plugin.js.map