"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.ChainConnectorDB = exports.uint32BE = exports.concatDBKeys = exports.checkDBError = exports.getDBInstance = void 0;
const lisk_sdk_1 = require("lisk-sdk");
const os = require("os");
const path_1 = require("path");
const fs_extra_1 = require("fs-extra");
const constants_1 = require("./constants");
const schemas_1 = require("./schemas");
const { Database } = lisk_sdk_1.db;
const getDBInstance = async (dataPath, dbName = 'lisk-framework-chain-connector-plugin.db') => {
    const dirPath = (0, path_1.join)(dataPath.replace('~', os.homedir()), 'plugins/data', dbName);
    await (0, fs_extra_1.ensureDir)(dirPath);
    return new Database(dirPath);
};
exports.getDBInstance = getDBInstance;
const checkDBError = (error) => {
    if (!(error instanceof lisk_sdk_1.db.NotFoundError)) {
        throw error;
    }
};
exports.checkDBError = checkDBError;
const concatDBKeys = (...keys) => Buffer.concat(keys);
exports.concatDBKeys = concatDBKeys;
const uint32BE = (val) => {
    const result = Buffer.alloc(4);
    result.writeUInt32BE(val, 0);
    return result;
};
exports.uint32BE = uint32BE;
class ChainConnectorDB {
    async load(dataPath) {
        this._db = await (0, exports.getDBInstance)(dataPath);
    }
    close() {
        this._db.close();
    }
    get privateKey() {
        return this._privateKey;
    }
    async saveToDBOnNewBlock(blockHeader) {
        const heightBuf = (0, exports.uint32BE)(blockHeader.height);
        const batch = new lisk_sdk_1.db.Batch();
        const newBlockHeaderBytes = lisk_sdk_1.codec.encode(schemas_1.blockHeaderSchemaWithID, blockHeader);
        batch.set((0, exports.concatDBKeys)(constants_1.DB_KEY_BLOCK_HEADER_BY_HEIGHT, heightBuf), newBlockHeaderBytes);
        if (!blockHeader.aggregateCommit.aggregationBits.equals(lisk_sdk_1.EMPTY_BYTES) ||
            !blockHeader.aggregateCommit.certificateSignature.equals(lisk_sdk_1.EMPTY_BYTES)) {
            const aggregateCommitHeight = (0, exports.uint32BE)(blockHeader.aggregateCommit.height);
            const aggregateCommitBytes = lisk_sdk_1.codec.encode(lisk_sdk_1.aggregateCommitSchema, blockHeader.aggregateCommit);
            batch.set((0, exports.concatDBKeys)(constants_1.DB_KEY_AGGREGATE_COMMIT_BY_HEIGHT, aggregateCommitHeight), aggregateCommitBytes);
        }
        await this._db.write(batch);
    }
    async getBlockHeaderByHeight(height) {
        try {
            const blockBytes = await this._db.get((0, exports.concatDBKeys)(constants_1.DB_KEY_BLOCK_HEADER_BY_HEIGHT, (0, exports.uint32BE)(height)));
            return lisk_sdk_1.codec.decode(schemas_1.blockHeaderSchemaWithID, blockBytes);
        }
        catch (error) {
            (0, exports.checkDBError)(error);
            return undefined;
        }
    }
    async getBlockHeadersBetweenHeights(fromHeight, toHeight) {
        const stream = this._db.createReadStream({
            gte: (0, exports.concatDBKeys)(constants_1.DB_KEY_BLOCK_HEADER_BY_HEIGHT, (0, exports.uint32BE)(fromHeight)),
            lte: (0, exports.concatDBKeys)(constants_1.DB_KEY_BLOCK_HEADER_BY_HEIGHT, (0, exports.uint32BE)(toHeight)),
            reverse: true,
        });
        const blockHeaders = await new Promise((resolve, reject) => {
            const list = [];
            stream
                .on('data', ({ value }) => {
                list.push(value);
            })
                .on('error', error => {
                reject(error);
            })
                .on('end', () => {
                resolve(list);
            });
        });
        return blockHeaders.map(b => lisk_sdk_1.codec.decode(schemas_1.blockHeaderSchemaWithID, b));
    }
    async deleteBlockHeadersBetweenHeight(fromHeight, toHeight) {
        const stream = this._db.createReadStream({
            gte: (0, exports.concatDBKeys)(constants_1.DB_KEY_BLOCK_HEADER_BY_HEIGHT, (0, exports.uint32BE)(fromHeight)),
            lte: (0, exports.concatDBKeys)(constants_1.DB_KEY_BLOCK_HEADER_BY_HEIGHT, (0, exports.uint32BE)(toHeight)),
            reverse: true,
        });
        const blockHeaderIndexes = await new Promise((resolve, reject) => {
            const list = [];
            stream
                .on('data', ({ key }) => {
                list.push(key);
            })
                .on('error', error => {
                reject(error);
            })
                .on('end', () => {
                resolve(list);
            });
        });
        const batch = new lisk_sdk_1.db.Batch();
        for (const key of blockHeaderIndexes) {
            batch.del(key);
        }
        await this._db.write(batch);
    }
    async deleteBlockHeaderByHeight(height) {
        const heightBuf = (0, exports.uint32BE)(height);
        await this._db.del((0, exports.concatDBKeys)(constants_1.DB_KEY_BLOCK_HEADER_BY_HEIGHT, heightBuf));
    }
    async getAggregateCommitByHeight(height) {
        try {
            const bytes = await this._db.get((0, exports.concatDBKeys)(constants_1.DB_KEY_AGGREGATE_COMMIT_BY_HEIGHT, (0, exports.uint32BE)(height)));
            return lisk_sdk_1.codec.decode(lisk_sdk_1.aggregateCommitSchema, bytes);
        }
        catch (error) {
            (0, exports.checkDBError)(error);
            return undefined;
        }
    }
    async getAggregateCommitBetweenHeights(fromHeight, toHeight) {
        const stream = this._db.createReadStream({
            gte: (0, exports.concatDBKeys)(constants_1.DB_KEY_AGGREGATE_COMMIT_BY_HEIGHT, (0, exports.uint32BE)(fromHeight)),
            lte: (0, exports.concatDBKeys)(constants_1.DB_KEY_AGGREGATE_COMMIT_BY_HEIGHT, (0, exports.uint32BE)(toHeight)),
            reverse: true,
        });
        const aggregateCommits = await new Promise((resolve, reject) => {
            const list = [];
            stream
                .on('data', ({ value }) => {
                list.push(value);
            })
                .on('error', error => {
                reject(error);
            })
                .on('end', () => {
                resolve(list);
            });
        });
        return aggregateCommits.map(a => lisk_sdk_1.codec.decode(lisk_sdk_1.aggregateCommitSchema, a));
    }
    async deleteAggregateCommitsBetweenHeight(fromHeight, toHeight) {
        const stream = this._db.createReadStream({
            gte: (0, exports.concatDBKeys)(constants_1.DB_KEY_AGGREGATE_COMMIT_BY_HEIGHT, (0, exports.uint32BE)(fromHeight)),
            lte: (0, exports.concatDBKeys)(constants_1.DB_KEY_AGGREGATE_COMMIT_BY_HEIGHT, (0, exports.uint32BE)(toHeight)),
            reverse: true,
        });
        const aggregateCommitIndexes = await new Promise((resolve, reject) => {
            const list = [];
            stream
                .on('data', ({ key }) => {
                list.push(key);
            })
                .on('error', error => {
                reject(error);
            })
                .on('end', () => {
                resolve(list);
            });
        });
        const batch = new lisk_sdk_1.db.Batch();
        for (const key of aggregateCommitIndexes) {
            batch.del(key);
        }
        await this._db.write(batch);
    }
    async deleteAggregateCommitByHeight(height) {
        const heightBuf = (0, exports.uint32BE)(height);
        await this._db.del((0, exports.concatDBKeys)(constants_1.DB_KEY_AGGREGATE_COMMIT_BY_HEIGHT, heightBuf));
    }
    async getValidatorsDataByHash(validatorsHash) {
        try {
            const bytes = await this._db.get((0, exports.concatDBKeys)(constants_1.DB_KEY_VALIDATORS_DATA_BY_HASH, validatorsHash));
            return lisk_sdk_1.codec.decode(schemas_1.validatorsDataSchema, bytes);
        }
        catch (error) {
            (0, exports.checkDBError)(error);
            return undefined;
        }
    }
    async setValidatorsDataByHash(validatorsHash, validatorsData, height) {
        const bytes = lisk_sdk_1.codec.encode(schemas_1.validatorsDataSchema, validatorsData);
        await this._db.set((0, exports.concatDBKeys)(constants_1.DB_KEY_VALIDATORS_DATA_BY_HASH, validatorsHash), bytes);
        await this._db.set((0, exports.concatDBKeys)(constants_1.DB_KEY_VALIDATORS_DATA_BY_HEIGHT, (0, exports.uint32BE)(height)), validatorsHash);
    }
    async getValidatorsDataByHeight(height) {
        try {
            const validatorHash = await this._db.get((0, exports.concatDBKeys)(constants_1.DB_KEY_VALIDATORS_DATA_BY_HEIGHT, (0, exports.uint32BE)(height)));
            return this.getValidatorsDataByHash((0, exports.concatDBKeys)(constants_1.DB_KEY_VALIDATORS_DATA_BY_HASH, validatorHash));
        }
        catch (error) {
            (0, exports.checkDBError)(error);
            return undefined;
        }
    }
    async deleteValidatorsHashByHeight(height) {
        const validatorHash = await this._db.get((0, exports.concatDBKeys)(constants_1.DB_KEY_VALIDATORS_DATA_BY_HEIGHT, (0, exports.uint32BE)(height)));
        await this.deleteValidatorsDataByHash((0, exports.concatDBKeys)(validatorHash));
    }
    async deleteValidatorsHashBetweenHeights(fromHeight, toHeight) {
        const stream = this._db.createReadStream({
            gte: (0, exports.concatDBKeys)(constants_1.DB_KEY_VALIDATORS_DATA_BY_HEIGHT, (0, exports.uint32BE)(fromHeight)),
            lte: (0, exports.concatDBKeys)(constants_1.DB_KEY_VALIDATORS_DATA_BY_HEIGHT, (0, exports.uint32BE)(toHeight)),
            reverse: true,
        });
        const validatorsHashes = await new Promise((resolve, reject) => {
            const list = [];
            stream
                .on('data', ({ key }) => {
                list.push(key);
            })
                .on('error', error => {
                reject(error);
            })
                .on('end', () => {
                resolve(list);
            });
        });
        for (const hash of validatorsHashes) {
            await this.deleteValidatorsDataByHash(hash);
        }
    }
    async deleteValidatorsDataByHash(validatorsHash) {
        const validatorData = await this.getValidatorsDataByHash(validatorsHash);
        if (!validatorData) {
            return;
        }
        await this._db.del((0, exports.concatDBKeys)(constants_1.DB_KEY_VALIDATORS_DATA_BY_HASH, validatorsHash));
        await this._db.del((0, exports.concatDBKeys)(constants_1.DB_KEY_VALIDATORS_DATA_BY_HEIGHT, (0, exports.uint32BE)(validatorData.height)));
    }
    async getAllValidatorsData() {
        const stream = this._db.createReadStream({
            gte: (0, exports.concatDBKeys)(constants_1.DB_KEY_VALIDATORS_DATA_BY_HASH, Buffer.alloc(4, 0)),
            lte: (0, exports.concatDBKeys)(constants_1.DB_KEY_VALIDATORS_DATA_BY_HASH, Buffer.alloc(4, 255)),
            reverse: true,
        });
        const validatorsData = await new Promise((resolve, reject) => {
            const list = [];
            stream
                .on('data', ({ value }) => {
                list.push(value);
            })
                .on('error', error => {
                reject(error);
            })
                .on('end', () => {
                resolve(list);
            });
        });
        return validatorsData.map(v => lisk_sdk_1.codec.decode(schemas_1.validatorsDataSchema, v));
    }
    async getCCMsByHeight(height) {
        const heightBuf = (0, exports.uint32BE)(height);
        let crossChainMessages = [];
        try {
            const encodedInfo = await this._db.get((0, exports.concatDBKeys)(constants_1.DB_KEY_CROSS_CHAIN_MESSAGES, heightBuf));
            crossChainMessages = lisk_sdk_1.codec.decode(schemas_1.ccmsAtHeightSchema, encodedInfo).ccms;
        }
        catch (error) {
            (0, exports.checkDBError)(error);
        }
        return crossChainMessages;
    }
    async deleteCCMsByHeight(height) {
        const heightBuf = (0, exports.uint32BE)(height);
        await this._db.del((0, exports.concatDBKeys)(constants_1.DB_KEY_CROSS_CHAIN_MESSAGES, heightBuf));
    }
    async getCCMsBetweenHeights(fromHeight, toHeight) {
        const stream = this._db.createReadStream({
            gte: (0, exports.concatDBKeys)(constants_1.DB_KEY_CROSS_CHAIN_MESSAGES, (0, exports.uint32BE)(fromHeight)),
            lte: (0, exports.concatDBKeys)(constants_1.DB_KEY_CROSS_CHAIN_MESSAGES, (0, exports.uint32BE)(toHeight)),
            reverse: true,
        });
        const ccmArrayAtEachHeight = await new Promise((resolve, reject) => {
            const list = [];
            stream
                .on('data', ({ value }) => {
                list.push(value);
            })
                .on('error', error => {
                reject(error);
            })
                .on('end', () => {
                resolve(list);
            });
        });
        const flatCCMList = [];
        for (const ccms of ccmArrayAtEachHeight) {
            flatCCMList.push(...lisk_sdk_1.codec.decode(schemas_1.ccmsAtHeightSchema, ccms).ccms);
        }
        return flatCCMList;
    }
    async deleteCCMsBetweenHeight(fromHeight, toHeight) {
        const stream = this._db.createReadStream({
            gte: (0, exports.concatDBKeys)(constants_1.DB_KEY_CROSS_CHAIN_MESSAGES, (0, exports.uint32BE)(fromHeight)),
            lte: (0, exports.concatDBKeys)(constants_1.DB_KEY_CROSS_CHAIN_MESSAGES, (0, exports.uint32BE)(toHeight)),
            reverse: true,
        });
        const ccmsListIndexes = await new Promise((resolve, reject) => {
            const list = [];
            stream
                .on('data', ({ key }) => {
                list.push(key);
            })
                .on('error', error => {
                reject(error);
            })
                .on('end', () => {
                resolve(list);
            });
        });
        const batch = new lisk_sdk_1.db.Batch();
        for (const key of ccmsListIndexes) {
            batch.del(key);
        }
        await this._db.write(batch);
    }
    async setCCMsByHeight(ccms, height) {
        const heightBuf = (0, exports.uint32BE)(height);
        const encodedInfo = lisk_sdk_1.codec.encode(schemas_1.ccmsAtHeightSchema, { ccms });
        await this._db.set((0, exports.concatDBKeys)(constants_1.DB_KEY_CROSS_CHAIN_MESSAGES, heightBuf), encodedInfo);
    }
    async getLastSentCCM() {
        let lastSentCCM;
        try {
            const encodedInfo = await this._db.get(constants_1.DB_KEY_LAST_SENT_CCM);
            lastSentCCM = lisk_sdk_1.codec.decode(schemas_1.lastSentCCMSchema, encodedInfo);
        }
        catch (error) {
            (0, exports.checkDBError)(error);
            return undefined;
        }
        return lastSentCCM;
    }
    async setLastSentCCM(ccm) {
        await this._db.set(constants_1.DB_KEY_LAST_SENT_CCM, lisk_sdk_1.codec.encode(schemas_1.lastSentCCMSchema, ccm));
    }
    async getListOfCCUs() {
        const stream = this._db.createReadStream({
            gte: (0, exports.concatDBKeys)(constants_1.DB_KEY_LIST_OF_CCU, Buffer.alloc(4, 0)),
            lte: (0, exports.concatDBKeys)(constants_1.DB_KEY_LIST_OF_CCU, Buffer.alloc(4, 255)),
            reverse: true,
        });
        const ccuBytes = await new Promise((resolve, reject) => {
            const list = [];
            stream
                .on('data', ({ value }) => {
                list.push(value);
            })
                .on('error', error => {
                reject(error);
            })
                .on('end', () => {
                resolve(list);
            });
        });
        return {
            list: ccuBytes
                .map(b => {
                const tx = lisk_sdk_1.codec.decode(schemas_1.transactionSchemaWithID, b);
                const txJSON = new lisk_sdk_1.chain.Transaction(tx).toJSON();
                const params = lisk_sdk_1.codec.decode(lisk_sdk_1.ccuParamsSchema, tx.params);
                const certificate = lisk_sdk_1.codec.decode(lisk_sdk_1.certificateSchema, params.certificate);
                const inboxUpdate = {
                    crossChainMessages: params.inboxUpdate.crossChainMessages.map(ccmBytes => {
                        const ccm = lisk_sdk_1.codec.decode(lisk_sdk_1.ccmSchema, ccmBytes);
                        return {
                            ...ccm,
                            fee: ccm.fee.toString(),
                            nonce: ccm.nonce.toString(),
                            params: ccm.params.toString('hex'),
                            receivingChainID: ccm.receivingChainID.toString('hex'),
                            sendingChainID: ccm.sendingChainID.toString('hex'),
                        };
                    }),
                    messageWitnessHashes: params.inboxUpdate.messageWitnessHashes.map(h => h.toString('hex')),
                    outboxRootWitness: {
                        bitmap: params.inboxUpdate.outboxRootWitness.bitmap.toString('hex'),
                        siblingHashes: params.inboxUpdate.outboxRootWitness.siblingHashes.map(h => h.toString('hex')),
                    },
                };
                const activeValidatorsUpdate = {
                    bftWeightsUpdate: params.activeValidatorsUpdate.bftWeightsUpdate.map(w => w.toString()),
                    bftWeightsUpdateBitmap: params.activeValidatorsUpdate.bftWeightsUpdateBitmap.toString('hex'),
                    blsKeysUpdate: params.activeValidatorsUpdate.blsKeysUpdate.map(key => key.toString('hex')),
                };
                return {
                    ...txJSON,
                    params: {
                        activeValidatorsUpdate,
                        certificate: {
                            aggregationBits: certificate.aggregationBits.toString('hex'),
                            blockID: certificate.blockID.toString('hex'),
                            height: certificate.height,
                            signature: certificate.signature.toString('hex'),
                            stateRoot: certificate.stateRoot.toString('hex'),
                            timestamp: certificate.timestamp,
                            validatorsHash: certificate.validatorsHash.toString('hex'),
                        },
                        inboxUpdate,
                        certificateThreshold: params.certificateThreshold.toString(),
                        sendingChainID: params.sendingChainID.toString('hex'),
                    },
                };
            })
                .sort((a, b) => Number(BigInt(b.nonce) - BigInt(a.nonce))),
            total: ccuBytes.length,
        };
    }
    async setCCUTransaction(ccu) {
        await this._db.set((0, exports.concatDBKeys)(constants_1.DB_KEY_LIST_OF_CCU, ccu === null || ccu === void 0 ? void 0 : ccu.id), lisk_sdk_1.codec.encode(schemas_1.transactionSchemaWithID, ccu));
    }
    async deleteCCUTransaction(ccuID) {
        await this._db.del((0, exports.concatDBKeys)(constants_1.DB_KEY_LIST_OF_CCU, ccuID));
    }
    async setPrivateKey(encryptedPrivateKey, password) {
        const parsedEncryptedKey = lisk_sdk_1.cryptography.encrypt.parseEncryptedMessage(encryptedPrivateKey);
        this._privateKey = Buffer.from(await lisk_sdk_1.cryptography.encrypt.decryptMessageWithPassword(parsedEncryptedKey, password, 'utf-8'), 'hex');
    }
    async deletePrivateKey(encryptedPrivateKey, password) {
        const parsedEncryptedKey = lisk_sdk_1.cryptography.encrypt.parseEncryptedMessage(encryptedPrivateKey);
        await lisk_sdk_1.cryptography.encrypt.decryptMessageWithPassword(parsedEncryptedKey, password, 'utf-8');
        this._privateKey = undefined;
    }
}
exports.ChainConnectorDB = ChainConnectorDB;
//# sourceMappingURL=db.js.map